# -*- coding: utf-8 -*-
"""
Author: Janos Toezser, Multi Channel Systems MCS GmbH
Date:   12/05/16
Status: Development
Description: Examplecode showing how to access the pipe generated by the
             CMOS-MEA-Control Server even on other PCs in network, how
             to decode the bytearray-data from the pipe into separate
             spikes and events and how to visualize them using 
             matplotlib and a GUI created with Tkinter. After starting 
             the server in CMOS-MEA-Control and entering/adjusting the 
             values in the entryfields of this PipeClient click the 
             "Start" button. A 2d-histogram will appear in the window 
             representig the spikes detected with CMOS-MEA-Control in 
             realtime. When selecting one of the displayed sensors with 
             the mouse-curser by left-clicking on it, the selected 
             sensor will be highlighted by a red rectangle and a second 
             plot will appear showing the detected waveforms of the 
             selected sensor. Repeated clicking on the 2d-histogram will 
             reset the highlighted sensor and the waveforms being 
             displayed to the newly selected sensor. Also data of 
             received events are displayed over a certain timeperiod in 
             a plot at the bottom of the window. Values for "Spike 
             Threshold", "Plotspeed" and "Timeframe" can be updated by 
             entering the new value, according to the types, and 
             pressing "Enter/Return". Types of displayed events can be
             selected and deselected with the checkboxes. When done, 
             stop the app by pressing the "Stop" button and then "Exit" 
             so the pipe gets closed correctly.  
"""

import os  # Used to set the path of the windowicon
import operator  # Used to sort lists of spikes/events by an attibute
import time  # needed to set some short breaks for processes to finish
import threading  # Generates threads for simultanious processes
import Tkinter as tk  # GUI 
import tkFont  # Change settings of fonts used by Tkinter
import tkMessageBox  # Errormessagebox
import Queue  # Queue for interal transmission and buffering of data
import struct  # Handles bytearrays received from CMOS-MEA-Control
import matplotlib as mpl  # Generates plots
mpl.use('TkAgg')  # Set matplotlib to work with Tkinter
import matplotlib.pyplot as plt  # Configure matplotlib to work with Tkinter
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg     
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
import numpy as np  # Needed for generation of colormap for lineplot


class Spike:
    """
    Spike-objects generated from pipe-data with informations about:
        SensorID, TimeStamp, plotTimer, Waveform
    """
    def __init__(self, SensorID, TimeStamp, Waveform=None):
        self.SensorID = SensorID
        self.TimeStamp = TimeStamp
        self.Waveform = Waveform
        

class Event:
    """
    Event-objects generated from pipe-data with informations about:
        source, ID, eventTimetamp, data, label
    """
    def __init__(self,source, ID, eventTimestamp, data, label):
        self.source = source
        self.ID = ID
        self.eventTimestamp = eventTimestamp
        self.data = data
        self.label = label


class App(tk.Frame):
    """
    GUI-object.
    Currently containes: 
    
    Bunch of variables for all sorts of functions, see inline comments.
    
    Figures f1,2,3:        
        Plots will be generated in these or the respective ax-subplots.
        
    Entryfields for: 
        - Hostname = PC-Name of computer in network which runs CMOS-MEA-
            Control Software as server (your PC = ".", use string)
        - Spike Threshold = Maximum number of spikes displayed per 
            Sensor (use ints)
        - Plotspeed = Refreshrate of plots (use floats)
        - Timeframe = Range of elapsed time from which spikes should be
            plotted (use floats)
    All entryfields except "Hostname" can be updated with immediate 
    effect while app is running by editing the value and pressing 
    "Return".
    
    Checkbuttons for Event-types:
        - Timer-Event,
        - DigitalIn-Event,
        - Stimulator-Event,
        - Application-Event
    Checkbuttons can be selected and deselected while app is running.                
    
    Buttons for:
        - "Start/Stop" = Start/Stop reading from pipe and decoding
            spikedata
        - "Exit" = Cleanup and exit
    "Exit" is best used when app is stopped. Also preferably use the 
    "Exit" button over the "X" of the window to avoid problems with 
    closing the pipe.
    
    Canvases for:
        - 2d histogram plot = Graphical representation of spikes
            detected on CMOS chip. Number of displayed spikes capped at
            "Spike Threshold" value. 
        - line plot = Graphical representation of waveforms for
            specified sensor on CMOS chip. Number of waveforms not
            capped. 
        - event plot = Graphical representation of events that occured
            within a given timeframe, divided into sourcetypes/IDs
    Sizes of canvases are fixed due to calculations of the sensor when
    clicking on plot. Can be changed but corresponding values have to be 
    adjusted...
    """
    def __init__(self, master):  
        """ 
            Initialize GUI-window.
            Generate entryfields, buttons, chechboxes, canvases for
            plots to be drawn in, textbox with event-info, a bunch of
            variables and lists which will be used later from
            memberfunctions, a pipe to get data from the server and 
            queues to safely get data from one function/thread to 
            another.
        """
        tk.Frame.__init__(self, master)
        self.pack()
 
        self.pipeToServer = ""  # Future pipe to server   
        self.queue = Queue.Queue()  # Internal queueu for data received from server
        self.textQueue = Queue.Queue() 
        self.histoList = []  # List of spike-objects for histogram plot
        self.lineList = []  # List of spikes-objects for line plot 
        self.eventList = []  # list of event-objects for event plot
        self.threadList = []  # List of threads to easily stop them when not needed
        self.eventTimestamp = 0  # Timestamp that gets sent by the server with timestamp event
        self.selectedSensor = 0  # Number of selected sensor to have its waveform plotted
        self.freeformCoords = (0,0)  # Sensormarker coordinates 
        self.freeformSize = 0  # 0 so sensormarker appears to be invisible 
        self.freeformText = ""  # Selectedsensor coords not displayed as no sensor selected yet
        self.appCheck = tk.IntVar()  # Variables for event-checkbuttons
        self.stimCheck = tk.IntVar()
        self.digCheck = tk.IntVar()
        self.timerCheck = tk.IntVar()
        self.isPlottingHisto = False  # Flags for plots of the respective type
        self.isPlottingLine = False
        self.isPlottingEvent = False
        self.isUpdatingText = False
        self.errorFlag = False  # Flag for errormessage
        self.errormessage = ""
                                               
        self.f1 = plt.figure(figsize=(6.2936507936507935, 6.0906298003072195))  # Adjusted for selected sensor x/y calculation.
        self.ax1 = self.f1.add_subplot(111)
        self.ax1.invert_yaxis()
        self.ax1.axis('off')
        
        self.f2 = Figure(figsize=(6.2936507936507935, 6.0906298003072195))  # Adjusted for symetry
        self.ax2 = self.f2.add_subplot(111)
        self.ax2.axis('off')
        
        self.f3 =Figure(figsize=(6.2936507936507935*2, 6.0906298003072195/2))  # Adjusted for symetry
        self.ax3 = self.f3.add_subplot(111)
        self.ax3.axis('off')
        
        self.topFrame = tk.Frame(master)  # Create parent frame and 3 child frames. 
        self.topFrame.pack() 
        self.entryFrame = tk.Frame(self.topFrame)  # Frame for lables, entryfields and buttons
        self.entryFrame.pack(side=tk.LEFT)
        self.eventPrintout = tk.Frame(self.topFrame)  # Frame for text about events
        self.eventPrintout.pack(side=tk.LEFT)
        self.eventFrame = tk.Frame(self.topFrame)  # Frame for event-canvas 
        self.eventFrame.pack(side=tk.BOTTOM)
        self.histoFrame = tk.Frame(self.topFrame)  # Frame for canvas #1
        self.histoFrame.pack(side=tk.LEFT)
        self.waveformFrame = tk.Frame(self.topFrame)  # Frame for canvas #2
        self.waveformFrame.pack()
        
        self.entryLabel = tk.Label(self.entryFrame, text="Client Settings")  # Label for settings area
        self.entryLabel.pack(side=tk.TOP,fill=tk.X, padx=8, pady=16)
        f = tkFont.Font(self.entryLabel, self.entryLabel.cget("font"))
        f.configure(underline = True)
        self.entryLabel.configure(font=f)
                                                        
        self.hostnameLabel = tk.Label(self.entryFrame, text="Hostname")  # Hostname can be '.' for own PC or actual name of PC in network
        self.hostnameLabel.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)  # String value required
        self.hostnameEntry = tk.Entry(self.entryFrame)
        self.hostnameEntry.insert(0,".")
        self.hostnameEntry.bind("<Return>", self.entry_update)  # Update Entryfieldvalues
        self.hostnameEntry.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)
        
        self.spikeThresholdLabel = tk.Label(self.entryFrame, text="Spike Threshold")  # Number of displayed spikes per sensor and timeframe
        self.spikeThresholdLabel.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)         # Integer value required     
        self.spikeThresholdEntry = tk.Entry(self.entryFrame)
        self.spikeThresholdEntry.insert(0,20)
        self.spikeThresholdEntry.bind("<Return>", self.entry_update)  # Update Entryfieldvalues
        self.spikeThresholdEntry.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)
        
        self.plotspeedLabel = tk.Label(self.entryFrame, text="Plotspeed in s")  # Speed with which plots will be updated, if possible 
        self.plotspeedLabel.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)        # Float value required
        self.plotspeedEntry = tk.Entry(self.entryFrame)
        self.plotspeedEntry.insert(0,0.2)
        self.plotspeedEntry.bind("<Return>", self.entry_update)  # Update Entryfieldvalues
        self.plotspeedEntry.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)

        self.timeframeLabel = tk.Label(self.entryFrame, text="Timeframe in s")  # Timeframe from which collected data will be displaed
        self.timeframeLabel.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)        # Float value required
        self.timeframeEntry = tk.Entry(self.entryFrame)
        self.timeframeEntry.insert(0,2.0)
        self.timeframeEntry.bind("<Return>", self.entry_update)  # Update Entryfieldvalues
        self.timeframeEntry.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)

        self.coordLabel = tk.Label(self.entryFrame, text="Selected sensor coords")  # Label for sensor coordinates
        self.coordLabel.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4) 
        
        self.coordsLabel = tk.Label(self.entryFrame, text="x: ,y: ")  # Coordinates of selected sensor will be displayed here
        self.coordsLabel.pack() 
        
        self.eventLabel = tk.Label(self.entryFrame, text="Event Selection")  # Label fpr events-area
        self.eventLabel.pack(side=tk.TOP, fill=tk.X, padx=8, pady=16)
        g = tkFont.Font(self.eventLabel, self.eventLabel.cget("font"))
        g.configure(underline = True)
        self.eventLabel.configure(font=g)
        
        self.eventText = tk.Text(self.eventPrintout, height=44, width=38)  # Textwindow for events  , yscrollcommand=self.eventScrollbar.set
        self.eventScrollbar = tk.Scrollbar(self.eventPrintout, command = self.eventText.yview)  # Scrollbar for event-text
        self.eventScrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.eventText.pack(padx=8, pady=4)
        self.eventText['yscrollcommand'] = self.eventScrollbar.set       
        
        self.checkFrame = tk.Frame(self.entryFrame)  # Frame for checkbuttons
        self.checkFrame.pack(side = tk.TOP)
        
        self.innerCheck_1 = tk.Frame(self.checkFrame)  # Internal splitting for button arrangement
        self.innerCheck_1.pack(side = tk.LEFT, padx=8)
        
        self.innerCheck_2 = tk.Frame(self.checkFrame)  # Internal splitting for button arrangement
        self.innerCheck_2.pack(side = tk.LEFT, padx=8)
        
        self.appCheckbutton = tk.Checkbutton(self.innerCheck_1)
        self.appCheckbutton.configure(text="App", variable = self.appCheck)  # Checkbutton for event selection
        self.appCheckbutton.pack(side = tk.TOP)
        self.stimCheckbutton = tk.Checkbutton(self.innerCheck_1)
        self.stimCheckbutton.configure(text="Stim", variable = self.stimCheck)
        self.stimCheckbutton.pack(side = tk.TOP, padx=8, pady=4)
        self.digCheckbutton = tk.Checkbutton(self.innerCheck_2)
        self.digCheckbutton.configure(text="DigIn", variable = self.digCheck)
        self.digCheckbutton.pack(side = tk.TOP)
        self.timerCheckbutton = tk.Checkbutton(self.innerCheck_2)
        self.timerCheckbutton.configure(text="Timer", variable = self.timerCheck)
        self.timerCheckbutton.pack(side = tk.TOP, padx=8, pady=4)       

        self.paddingLabel1 = tk.Label(self.entryFrame, height=12, text = " ")  # Just a spacer
        self.paddingLabel1.pack(side = tk.TOP, padx=8, pady=4)
                
        self.button1 = tk.Button(self.entryFrame)  # Start-/Stop-button. Will be specified later in functions self.resetbutton and self.startthread
        self.button1.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)

        self.button2 = tk.Button(self.entryFrame)  # Exit-button ...
        self.button2.config(text="Exit", command=self.exiter)
        self.button2.pack(side=tk.TOP, fill=tk.X, padx=8, pady=4)
                        
        self.waveformCanvas = tk.Canvas(self.waveformFrame)  # Waveforms of spike will be drawn here
        self.waveformCanvas.pack()
                             
        self.histoCanvas = tk.Canvas(self.histoFrame)  # Histogram of chip will be drawn here
        self.histoCanvas.pack()
        
        self.eventCanvas = tk.Canvas(self.eventFrame)  # Events will be drawn here
        self.eventCanvas.pack()
                      
        self.canvas2 = FigureCanvasTkAgg(self.f2, master=self.waveformCanvas)  # Setting canvas properties for matplotlib
        self.plotCanvas2 = self.canvas2.get_tk_widget()
        self.plotCanvas2.pack(side=tk.RIGHT, fill=tk.BOTH, expand=1)  
               
        self.canvas1 = FigureCanvasTkAgg(self.f1, master=self.histoCanvas)  # Setting canvas properties for matplotlib
        self.plotCanvas1 = self.canvas1.get_tk_widget()
        self.plotCanvas1.bind("<Button-1>", self.select_sensor)  # Bind mousclick-event to histogram-canvas
        self.plotCanvas1.pack(side=tk.LEFT, fill=tk.BOTH, expand=1)  
        
        self.canvas3 =  FigureCanvasTkAgg(self.f3, master=self.eventCanvas)  # Setting canvas properties for matplotlib
        self.plotCanvas3 = self.canvas3.get_tk_widget()
        self.plotCanvas3.pack(fill=tk.BOTH, expand=1)
        
        self.resetbutton()  # calling the function resetbutton() upon start to set "Start" button functions
                
     
    def resetbutton(self):  
        """ 
            Set flag for threads to False. Deactivate running threads if
            any active. Reset command and text on button1 and some other 
            variables. 
        """
        self.running = False  # Stop threads by setting flag to False
        try:  # Try to deactivate any unwanted running threads
            for thread in self.threadList:
                thread.run = False
        except Exception:
            pass
        time.sleep(0.1)  # Allow processes to finish
        if self.pipeToServer != "":
            while not self.pipeToServer.closed:
                self.pipeToServer.close()
            self.pipeToServer = ""
        with self.textQueue.mutex:
            self.textQueue.queue.clear()    
        self.freeformSize = 0  # Reset freeformsize so no longer displayed
        self.coordsLabel.config(text="")  # Reset x/y-coords displayed to nothing
        self.selectedSensor = 0
        self.button1.config(text="Start", command=self.start_thread)  # Change text and function of button1

    
    def start_thread(self):  
        """ 
            Check for correctness of entered values. Set flag for 
            threads, self.running, on True. Get values from entryfields.
            Check if all the required data for the subsequent plot-
            functions are met, namely the entryfields hold sufficient 
            values. Open pipe with the self.hand_shake() function and 
            send handshake. Start threads to read from pipe and to 
            decode info sent by CMOS-MEA-Control. Threads are required 
            so this processes never cease and also doesn't threads will 
            not cause deadlocks on the other functions as regular while 
            loops would. Start recurring plot function calls with the 
            built in tk.after() method. Flags are set so that multiple 
            start/stops of the App don't start multiple plot- functions.
            This method has been chosen as using threads proved to 
            severely reduce performance. At the end set function and 
            label of the button back to "Stop" and resetbutton(). 
        """
        self.value_validator()
        
        if  not self.errorFlag:  
            self.hostname = self.hostnameEntry.get()  
            self.spikethreshold = int(self.spikeThresholdEntry.get())
            self.plotspeed = float(self.plotspeedEntry.get())
            self.timeframe = (float(self.timeframeEntry.get()))*(10**6)
            if self.pipeToServer == "":
                self.hand_shake()
            self.running = True
            readpipeThread = threading.Thread(target=self.read_pipe)
            self.threadList.append(readpipeThread)
            readpipeThread.start()
            readqueueThread = threading.Thread(target=self.read_queue)
            self.threadList.append(readqueueThread)
            readqueueThread.start()
            time.sleep(0.1)
            
            if self.isPlottingHisto == False:
                self.isPlottingHisto = True
                self.histoStarter()
                
            if self.isPlottingLine == False:
                self.isPlottingLine = True
                self.lineStarter()
                
            if self.isPlottingEvent == False:
                self.isPlottingEvent = True
                self.eventStarter()
            
            if self.isUpdatingText == False:
                self.isUpdatingText = True
                self.textStarter()    
            
            self.button1.config(text="Stop", command=self.resetbutton)  
        else:
            tkMessageBox.showerror("Error",self.errormessage)
        self.errorFlag = False
        self.errormessage = ""
    
    
    def value_validator(self):
        """
            Checks if values in entryfields are of correct format. If 
            not, error-flags are set to True and the error-message that
            will be displayed in error-window is updated with info 
            about the errors that occured. 
        """
        threshError = False
        speedError = False
        timeError = False
        if len(self.hostnameEntry.get()) == 0:
            self.errorFlag = True
            self.errormessage += "Hostname empty \n"
        if len(self.spikeThresholdEntry.get()) == 0:        
            self.errorFlag = True
            self.errormessage += "Threshold empty \n"
        if len(self.plotspeedEntry.get()) == 0:    
            self.errorFlag = True
            self.errormessage += "Plotspeed empty \n"
        if len(self.timeframeEntry.get()) == 0:
            self.errorFlag = True
            self.errormessage += "Timeframe empty \n"
        values = ["0","1","2","3","4","5","6","7","8","9","0","."]
        threshCheck = self.spikeThresholdEntry.get()
        for i in xrange(len(threshCheck)):
            if threshCheck[i] not in values and threshError == False:
                self.errormessage += "Unsupported value in Threshold \n"
                threshError = True
                self.errorFlag = True
        speedCheck = self.plotspeedEntry.get()
        for i in xrange(len(speedCheck)):
            if speedCheck[i] not in values and speedError == False:
                self.errormessage += "Unsupported value in Plotspeed \n"
                speedError = True
                self.errorFlag = True
        timeCheck = self.timeframeEntry.get()
        for i in xrange(len(timeCheck)):    
            if timeCheck[i] not in values and timeError == False:
                self.errormessage += "Unsupported value in Timeframe \n"
                timeError = True
                self.errorFlag = True
        threshError = False
        speedError = False
        timeError = False        
            
                                        
    def entry_update(self, event):
        """ 
            Check for correctness of entered values and shows an error-
            window with info about the errors that occured. Get updated
            values from inputfields. 
        """
        self.value_validator()
        
        if not self.errorFlag:
            self.hostname = self.hostnameEntry.get()  
            self.spikethreshold = int(self.spikeThresholdEntry.get())
            self.plotspeed = float(self.plotspeedEntry.get())
            self.timeframe = (float(self.timeframeEntry.get()))*(10**6)
        else:
            tkMessageBox.showerror("Error",self.errormessage)
        self.errorFlag = False
        self.errormessage = ""


    def text_box_update(self):
        """
            Grab text of current event from queue if there is any in 
            self.textQueue and update text displayed in textbox but only
            if corresponding checkbutton is selected. Textupdate can be
            laggy if large amounts of spikes/events are received/
            displayed. Timerevent best remains deactivated when other 
            plots start slowing down.
        """
        if self.running:
            if not self.textQueue.empty():
                text = self.textQueue.get()
                if self.timerCheck.get() == 1: 
                    if text[0] == 4:
                        self.eventText.insert(tk.END, text[1])
                        
                if self.digCheck.get() == 1:
                    if text[0] == 3:
                        self.eventText.insert(tk.END, text[1])
                        
                if self.stimCheck.get() == 1:
                    if text[0] == 2:
                        self.eventText.insert(tk.END, text[1])
                        
                if self.appCheck.get() == 1:    
                    if text[0] == 1:
                        self.eventText.insert(tk.END, text[1])
                self.eventText.see(tk.END)
                      
    
    def textStarter(self):
        """
            Recurring function that calls itself and the self.tex_box_
            update() function. With .after we restart the function that
            starts the update function in dependence of plotspeed. In
            the end the textbox will be updated with info about the 
            latest event that got received. 
        """
        self.text_box_update()
        self.master.after(int(self.plotspeed*5),self.textStarter) 
    
    
    def histoStarter(self):
        """
            Recurring function that calls itself and self.histo_plot()
            function. With .after we restart the function that starts
            the plot function in dependence of plotspeed. Ultimately the
            2d-representation of the CMOS chip will be plotted.
        """
        self.histo_plot()
        self.master.after(int(self.plotspeed*1000),self.histoStarter) 

        
    def lineStarter(self):
        """
            Recurring function that callsitself and self.line_plot()
            function. With .after we restart the function that starts
            the plot function in dependence of plotspeed. At last
            waveforms will be plotted for the sensor selected in the
            2d-CMOS plot.
        """
        self.line_plot()
        self.master.after(int(self.plotspeed*1000),self.lineStarter) 

                
    def eventStarter(self):
        """
            Recurring function that callsitself and self.event_plot()
            function. With .after we restart the function that starts
            the plot function in dependence of plotspeed. In the end
            events will be plotted.
        """
        self.event_plot()
        self.master.after(int(self.plotspeed*1000),self.eventStarter)              
 
                                                                                                       
    def hand_shake(self):  
        """ 
            Generate handshakemessage as bytearray following the
            instructions of the CMOS-MEA-Control programming guide, open
            named pipe, a shared-memory-segment, to CMOS-MEA-Control and
            send handshake so server starts providing data. 
        """
        signature = bytearray("McsCMOSClient")  # Generate handshake message   
        size = struct.pack("i", len(signature)*2)
        newSignature = bytearray()
        for i in xrange(len(signature)):
            newSignature.append(signature[i])
            newSignature.append(0)
        handshake = struct.pack("i", 1) + size + size + newSignature
        pipename = r'\\'+self.hostname+'\pipe\McsCMOSNamedPipe'
        self.pipeToServer = open(pipename, 'r+b', 0)  # Open pipe
        self.pipeToServer.write(handshake)  # Send handshake message  


    def read_pipe(self):  
        """ 
            Read from named pipe. Determine DataType and put data + info
            about size and type in internal queue. The queue is needed
            as data gets shuffled between threads and rules of thread-
            safety need to be obeyed.
            
            Datatypes encoded in header can be:
                - 1: Server-Handshake
                - 2: Events
                - 3: Spikes
                
            When additional DataTypes get added to the Server Software
            they can easily be impemented here.
        """
        while self.running:
            packetHeader = self.pipeToServer.read(12)  # Get and decode header
            DataType = struct.unpack("i", packetHeader[0:4])[0]
                            
            if DataType == 3:  # Spikedata
                blocksize = struct.unpack("i", packetHeader[4:8])[0]
                itemsize = struct.unpack("i", packetHeader[8:12])[0]
                datablock = (3,itemsize, self.pipeToServer.read(blocksize))
                self.queue.put(datablock)  # Provide spikedata to other thread

            elif DataType == 2:  # Eventdata
                blocksize = struct.unpack("i", packetHeader[4:8])[0]
                itemsize = struct.unpack("i", packetHeader[8:12])[0]
                datablock = (2, blocksize, self.pipeToServer.read(blocksize))
                self.queue.put(datablock)  # Provide eventdata to other function
                
            elif DataType == 1:  # Server Handshake
                blocksize = struct.unpack("i", packetHeader[4:8])[0] 
                itemsize = struct.unpack("i", packetHeader[8:12])[0]
                datablock = self.pipeToServer.read(blocksize)


    def read_queue(self):  
        """ 
            Read from queue and generate instances of Spike- or Event-
            objects to be used for plots. Also generate text with event-
            info for textbox.
            Spike- and Event-objects can be stored in lists which are 
            accessed by the plotting functions. The text has to be put 
            into a queue due to threadsafety reasons.
        """
        while self.running:
            currentBlock = self.queue.get()  # Get block of data from queue 
            
            if currentBlock[0] == 2:  # If currentBlock is an event, decode and generate an event-object
                blocksize = currentBlock[1]
                datablock = currentBlock[2]
                source = struct.unpack("i", datablock[:4])[0]
                self.source = source
                ID = struct.unpack("i", datablock[4:8])[0]
                eventTimestamp = struct.unpack("Q", datablock[8:16])[0] 
                data = struct.unpack("I", datablock[16:20])[0]
                labelLength = "s"*(blocksize-20)
                label = struct.unpack(labelLength, datablock[20:(20+blocksize)])
                
                newLabel = ""
                for letter in xrange(0,len(label),2):
                    newLabel += label[letter]

                if source == 4:  # If event is TimerEvent update self.eventTimestamp
                    self.eventTimestamp = eventTimestamp
                    
                event = Event(source, ID, eventTimestamp, data, label)
                self.eventList.append(event)  # Add event to eventList for plotting
                
                nowTime = str(time.strftime("%H:%M:%S"))
                eventInfo = nowTime+": "+newLabel+"_"+str(ID)+" at "+str(eventTimestamp)+" \n"
                text = (source,eventInfo)
                
                self.textQueue.put(text)
                        
            elif currentBlock[0] == 3:  # If currentBlock is a spike, decode and generate a spike-object
                itemsize = currentBlock[1]
                datablock = currentBlock[2]
                itemcount = len(datablock)/itemsize  # Calculate number of included spikes
                
                for item in xrange(itemcount):  # Decode spike-info     
                    SensorID = struct.unpack("i", datablock[:4])[0]  # Extract SensorID, TimeStamp  
                    TimeStamp = struct.unpack("q", datablock[4:12])[0]  
                    if itemsize > 12:  # Check if Waveform is included. If so, extract it
                        lenwave = "d" * ((itemsize-12) / 8)
                        Waveform = struct.unpack(lenwave, datablock[12:itemsize])
                        spike = Spike(SensorID, TimeStamp, Waveform)
                    else:  # If not, don't ...
                        spike = Spike(SensorID, TimeStamp)
                    datablock = datablock[itemsize:]
                    self.histoList.append(spike)
                    self.lineList.append(spike)
                    
            elif currentBlock[0] == 4:  # Reserved for future data-types
                print("Mystery Data Received!")  # Just kidding! We know what we are doing ... most of the time ...
                   
    
    def histo_plot(self):  
        """ 
            Draw a histogram representing detected spikes on CMOS-chip.
            First a copy of the current list with spike-objects for this
            plot is copied so iterating over the objects is possible 
            while the original list keeps on receiving additional spikes.
            The list is sorted according to the 'TimeStamp' of the 
            spikes so that after the plot is finished, when updating the
            original list, the spikes which are "older" than the 
            selected timeframe are removed. This is done to keep the 
            list as short and the performance as good as possible. For 
            this purpose the rangevariables are introduced in the 
            beginning. The x/y-coordinates are calculated with the 
            SensorID. In the first column there are sensors 1 to 65 in 
            the second 66 to 130 and so on up to 65*65 = 4225. Plot is 
            rater basic. A freeform is added upon mouse click to visually 
            highlight the selected sensor which will get its waveforms 
            plotted in the line_plot() function. In the end, the 
            self.histoList will be updated with by replacing the portion 
            of the list, which existed before the plot began with the 
            portion of the copied list that contains all the spikes 
            "young" enough to be plotted.  
        """
        histoCopy = list(self.histoList)
        histoCopy.sort(key=operator.attrgetter('TimeStamp'))  # Sort list by "TimeStamp"        
        sensorlist = []  # Reset needed lists for plot
        xlist = []
        ylist = []
        rangeafterplot = 0  # Reset counter of plotted spikes   
        rangebeforeplot = len(histoCopy)  # Get value for iteration
                
        for spike in histoCopy:  # Select Spikes to be plotted
            if float(self.eventTimestamp - spike.TimeStamp) < self.timeframe:  # Check "age" of spike
                rangeafterplot += 1
                currentsensorid = spike.SensorID
                if sensorlist.count(currentsensorid) < int(self.spikethreshold)+1:  # limit number of displayed spikes per sensor 
                    sensorlist.append(currentsensorid)
                    if currentsensorid % 65 == 0:  # Convert Sensors to x/y coords
                        xlist.append(currentsensorid / 65) 
                        ylist.append(65)
                    else:
                        xlist.append((currentsensorid / 65)+1)
                        ylist.append((currentsensorid % 65))
                    
        plt.set_cmap("Greys")  # Set plot properties, generate plot, draw plot
        plt.clf()  # Clear plot
        plt.hist2d(xlist, ylist, bins=(xrange(1,67,1),xrange(1,67,1)))  # Create plot
        currentaxis = plt.gca()
        currentaxis.invert_yaxis()  # y-axis inverted to match CMOS-MEA-Control display
        currentaxis.add_patch(Rectangle(self.freeformCoords, self.freeformSize, self.freeformSize, facecolor="red"))  # add red rectangle
        plt.axis('off')  # Don't display axis
        plt.gcf()
        self.canvas1.draw()  # Draw plot
                        
        disposable = rangebeforeplot-rangeafterplot  # Trim original list
        self.histoList[:rangebeforeplot] = histoCopy[disposable:]  
 

    def select_sensor(self, event):  
        """ 
            On left-click on chip-plot retrieve cursercoordinates and
            calculate SensorID that gets plotted by self.line_plot().
            Values have to be adjusted as plot is not 100% symmetrical. 
        """
        x = event.x  # Get x/y coordinates of mouse-curser
        y = event.y  
        x_min = 63  # Pixel offset of plot.
        y_min = 48  
        stepSize = 6  # Pixel-width of sensors on plot
        xCoords = ((x - x_min)/stepSize)  # Calculate actual x/y coordinates  
        yCoords = ((y - y_min)/stepSize)  
        
        self.selectedSensor = (((xCoords*65) + (yCoords+1)))  # Set selected Sensor  
                
        if x <= 34:  # Calculate offset adjusted x/y coordinates for sensor-highlight
            if y <= 34:   
                xCoordsNew = xCoords + 0.9   
                yCoordsNew = yCoords + 1.15
            elif y >= 37:
                xCoordsNew = xCoords + 0.9   
                yCoordsNew = yCoords + 0.85
        elif x >= 37:
            if y <= 34:   
                xCoordsNew = xCoords + 1.1   
                yCoordsNew = yCoords + 1.15
            elif y >= 37:
                xCoordsNew = xCoords + 1.1  
                yCoordsNew = yCoords + 0.85
        self.freeformCoords = (xCoordsNew, yCoordsNew)  # Set highlight coordinates
        self.freeformSize = 1  # Change size of sensormarker to 1 so it is visible
        if xCoords >= 0 and xCoords <= 64 and yCoords >= 0 and yCoords <= 64:
            self.freeformText = "x: "+str(xCoords + 1)+", y: "+str(yCoords + 1)  # Set text with x/y coordinates
        else:
            self.freeformText = ""  # Nothing is displayed when clicked outside of plot
        self.coordsLabel.config(text=self.freeformText)  # display x/y coordinates

            
    def line_plot(self):  
        """ 
            Draw lines for waveforms of Spikes matching the selected
            SensorID.First a copy of the current list with spike-objects
            for this plot is made so iterating over the objects is
            possible. The list is sorted according to the 'TimeStamp' of
            the spikes so that after the plot is finished, when updating
            the original list, the spikes which are "older" than the 
            selected timeframe are removed. This is done to keep the 
            list as short and the performance as good as possible. For 
            this purpose the rangevariables are introduced in the 
            beginning. Next the waveforms are extracted but only for the 
            spikes that where detected by the sensor with the matching 
            ID to the selected sensor. Plot is rather basic but custom 
            colormap is generated so that the youngest waveforms are 
            displayed in black and the older they get the lighter/greyer 
            they turn. In the end, the self.lineList will be updated 
            with by replacing the portion of the list, which existed 
            before the plot began with the portion of the copied list 
            that contains all the spikes "young" enough to be plotted. 
        """        
        lineCopy = list(self.lineList)
        lineCopy.sort(key=operator.attrgetter('TimeStamp'))  # Sort list by "TimeStamp"
        wavelist = []  # Reset list of plotted items
        rangeafterplot = 0  # Reset counter for plotted spikes
        rangebeforeplot = len(lineCopy)  # Get value for iteration
        
        for spike in lineCopy:  # Select Waveforms to be plotted
            if float(self.eventTimestamp - spike.TimeStamp) < self.timeframe:  # Check "age" of spike
                rangeafterplot += 1
                if spike.SensorID == self.selectedSensor:  # Check if SensorId matches selected Sensor
                    currentWaveform = spike.Waveform
                    wavelist.append(currentWaveform)  # Add waveform to list of plotted items
                                    
        color=iter(plt.cm.Greys(np.linspace(0,1,len(wavelist))))  # Create custom colormap
        self.ax2.cla()  # Clear plot
        for i in xrange(len(wavelist)):  # Plot collected waveforms in one plot
            c = next(color)
            self.ax2.plot(wavelist[i], c = c)
        self.canvas2.draw()  # Draw plot
                    
        disposable = rangebeforeplot - rangeafterplot  # Trim original list
        self.lineList[:rangebeforeplot] = lineCopy[disposable:]  
                
        
    def event_plot(self):
        """ 
            Draw events as they occur over time for a given timeframe.
            First a copy of the current list with event-objects is
            made so iterating over the objects is possible. The list is 
            sorted according to the 'eventTimeStamp' of the events so 
            that after the plot is finished, when updating the original 
            list, the events which are "older" than the selected 
            timeframe are removed. This is done to keep the list as 
            short and the performance as good as possible. For this
            purpose the rangevariables are introduced in the beginning. 
            First the highest number of IDs for every event type is 
            detected, ___Max. From these values the total maximum of IDs 
            is calculated, yMax. This value will be used later on to 
            dynamically calculate the y-values of the different events. 
            Next, a list of lists is generated for every event source/ID 
            combination. These lists are filled in the next step with 
            the respective eventTimeStamps. Plot is a bit tricky. Only 
            those events with active checkbuttons will get plotted. For
            displaying reasons some adjustments are made, altering the 
            standard settings of the plot. In the end, the 
            self.eventList will be updated with by replacing the portion 
            of the list, which existed before the plot began with the 
            portion of the copied list that contains all the events 
            "young" enough to be plotted.
        """
        eventCopy = list(self.eventList)  
        eventCopy.sort(key=operator.attrgetter('eventTimestamp'))  
        yMax = 0  
        appMax = 0  
        stimMax = 0
        digMax = 0
        timerMax = 0
        eventCounter = 0  
        rangeafterplot = 0  # Reset counter of plotted events   
        rangebeforeplot = len(eventCopy)  # Get value for iteration
                
        appEventPlotList = []  # Create lists for lists for each event type
        stimEventPlotList = []
        digInEventPlotList = []
        timerEventPlotList = []
        
        for event in eventCopy:  
            if event.source == 1:
                if event.ID > appMax:
                    appMax = event.ID
            elif event.source == 2:
                if event.ID > stimMax:
                    stimMax = event.ID
            elif event.source == 3:
                if event.ID > digMax:
                    digMax = event.ID
            elif event.source == 4:
                if event.ID > timerMax:
                    timerMax = event.ID
        
        yMax = appMax + stimMax + digMax + timerMax  
       
        for ID in xrange(appMax):  # Create lists for lists for each ID instance of event types
            appEventPlotList.append([]) 
        for ID in xrange(stimMax):
            stimEventPlotList.append([])
        for ID in xrange(digMax):
            digInEventPlotList.append([])
        for ID in xrange(timerMax):
            timerEventPlotList.append([])
            
        for event in eventCopy:
            timestamp = event.eventTimestamp
            if float(self.eventTimestamp - timestamp) < 1.5*10**7:  # Check "age" of event
                if event.source == 1:  # Append events in corresponding event/ID lists
                    rangeafterplot += 1
                    appEventPlotList[event.ID-1].append(timestamp/10.0**6)
                elif event.source == 2:
                    rangeafterplot += 1
                    stimEventPlotList[event.ID-1].append(timestamp/10.0**6)
                elif event.source == 3:
                    rangeafterplot += 1
                    digInEventPlotList[event.ID-1].append(timestamp/10.0**6)
                elif event.source == 4:
                    rangeafterplot += 1
                    timerEventPlotList[event.ID-1].append(timestamp/10.0**6)
                else:
                    print("Unknown Event!")  
                        
        self.ax3.cla()  # Plot events over time according to EventSource 
        if self.timerCheck.get() == 1:  # But only when checkbutton is selected
            for sublist in timerEventPlotList:
                labelText = "Timer_"+str(timerEventPlotList.index(sublist)+1)
                yValues = [(yMax-eventCounter) for event in sublist]  # y axis generated with yMax for each x value with list comprehension
                self.ax3.plot(sublist, yValues, '^', label=labelText)  # x value is timestamp of event,  
                eventCounter += 1 
        if self.digCheck.get() == 1:
            for sublist in digInEventPlotList:
                labelText = "DigIn_"+str(digInEventPlotList.index(sublist)+1)
                yValues = [(yMax-eventCounter) for event in sublist]
                self.ax3.plot(sublist, yValues, '^', label=labelText)     
                eventCounter += 1
        if self.stimCheck.get() == 1:
            for sublist in stimEventPlotList:
                labelText = "Stim_"+str(stimEventPlotList.index(sublist)+1)
                yValues = [(yMax-eventCounter) for event in sublist]
                self.ax3.plot(sublist, yValues, '^', label=labelText)   
                eventCounter += 1
        if self.appCheck.get() == 1:
            for sublist in appEventPlotList:
                labelText ="App_"+str(appEventPlotList.index(sublist)+1)
                yValues = [(yMax-eventCounter) for event in sublist]
                self.ax3.plot(sublist, yValues, '^', label=labelText)    
                eventCounter += 1  
        self.ax3.set_xlim([((self.eventTimestamp-(1.5*10**7))/10.0**6),(self.eventTimestamp/10.0**6)])  # set range of values used as x-axis
        self.ax3.set_ylim([0,(yMax+1)])  # set range of values of y-axis
        self.ax3.yaxis.set_visible(False)  # turn y-axis off so no colission with plot legend
        self.ax3.set_yticks(np.arange(0,yMax+1,1))  # set step-size of y-axis 
        self.ax3.legend(bbox_to_anchor=(-0.165, 1.06), loc = "upper left", frameon = False)  # legend of events 
        self.canvas3.show()
        
        disposable = rangebeforeplot-rangeafterplot  # Trim original list
        self.eventList[:rangebeforeplot] = eventCopy[disposable:]
        
                                                                 
    def exiter(self):  
        """ 
            On buttonpress turn threads of, close pipe and quit. 
        """
        self.running = False 
        if self.pipeToServer != "":
            self.pipeToServer.close()
        self.quit()

            
root = tk.Tk()
pti = os.path.dirname(os.path.realpath(__file__))  # Get path of window-icon. Icon needds to be in folder of script
root.wm_iconbitmap(r''+pti+'\CMC.ico')  # Set icon of window 
root.wm_title("Python PipeClient")  # Set title of Window
app = App(root) 
app.mainloop()
root.destroy()        
